/**************************************************************************
 *  This file is part of QXmlEdit                                         *
 *  Copyright (C) 2011 by Luca Bellonda and individual contributors       *
 *    as indicated in the AUTHORS file                                    *
 *  lbellonda _at_ gmail.com                                              *
 *                                                                        *
 * This library is free software; you can redistribute it and/or          *
 * modify it under the terms of the GNU Library General Public            *
 * License as published by the Free Software Foundation; either           *
 * version 2 of the License, or (at your option) any later version.       *
 *                                                                        *
 * This library is distributed in the hope that it will be useful,        *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU      *
 * Library General Public License for more details.                       *
 *                                                                        *
 * You should have received a copy of the GNU Library General Public      *
 * License along with this library; if not, write to the                  *
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,       *
 * Boston, MA  02110-1301  USA                                            *
 **************************************************************************/

#include "compareDialog.h"
#include "ui_Compare.h"
#include "utils.h"
#include "compare.h"
#include <QHash>

#define FILE_TEST   "/home/lbell/devel/qxmledit/qxmledit/src/data/compare_to.xml"

#define CSS_TEXT    "\
<style type='text/css'>\
body {                  \
   font-size: 10pt;     \
   background-color:#FFF;\
}                       \
.T0 {                   \
    color: #888;        \
    font-style:italic;  \
}                      \
.T1 {                   \
    color: #F44;        \
    font-style:italic;  \
}                      \
.E0 {                   \
    color: #44F;        \
}                      \
.E1 {                   \
    color: #888;        \
}                      \
.E2 {                   \
    color: #F42;        \
}                      \
.E3 {                   \
    color: #0A0;        \
}                      \
.A0 {                   \
    color: #44F;        \
}                      \
.A1 {                   \
    color: #888;        \
}                      \
.A2 {                   \
    color: #F42;        \
}                      \
.A3 {                   \
    color: #0A0;        \
}                      \
.C0 {                   \
    color: #CCF;        \
    font-style:italic;  \
}                      \
.C1 {                   \
    color: #CCC;        \
    font-style:italic;  \
}                      \
.C2 {                   \
    color: #FCC;        \
    font-style:italic;  \
}                      \
.C3 {                   \
    color: #CFC;        \
    font-style:italic;  \
}                      \
.AV {                   \
    color: #BB92C5;        \
}                      \
</style>"


void Compare::doCompare(QWidget *parent, Regola *original)
{
    Compare compare(original, parent);
    compare.setModal(true);
    if(!compare.loadData()) {
        return ;
    }
    compare.exec();
}

//----------------------------------------------------------------------


Compare::Compare(Regola* newOriginal, QWidget *parent) :
    QDialog(parent),
    ui(new Ui::Compare),
    _areDifferent(false)
{
    toCompare = NULL ;
    original = newOriginal ;
    ui->setupUi(this);
}

Compare::~Compare()
{
    if(NULL != toCompare) {
        delete toCompare ;
    }
    delete ui;
}

bool Compare::loadData()
{
#ifdef DEBUG_COMPARE
    compare.loadRule(FILE_TEST);
#else
    QString filePath = QFileDialog::getOpenFileName(parentWidget(),  tr("Open File To Compare to"),
                       original->fileName(), tr("XML files (*.xml);;All files (*);;"));
    if(!filePath.isEmpty()) {
        loadRule(filePath);
    } else {
        return false ;
    }
#endif
    return true;
}

bool Compare::loadData(const QString &filePath)
{
    if(!filePath.isEmpty()) {
        loadRule(filePath);
    } else {
        return false ;
    }
    return true;
}

bool Compare::areDifferent()
{
    return _areDifferent ;
}
void Compare::changeEvent(QEvent *e)
{
    QDialog::changeEvent(e);
    switch(e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

bool Compare::loadFile(const QString &filePath)
{
    bool isOk = false;
    if(!filePath.isEmpty()) {
        QFile file(filePath);
        if(file.open(QIODevice::ReadOnly)) {
            if(document.setContent(&file)) {
                isOk = true;
            } else {
                Utils::error(tr("Unable to parse XML"));
            }
            file.close();
        } else {
            Utils::error(QString(tr("Unable to load file.\n Error code is '%1'")).arg(file.error()));
        }
    } else {
        Utils::error(tr("File name empty. Unable to load it."));
    }
    return isOk;
}

bool Compare::loadData(QByteArray dataIn)
{
    QBuffer file(&dataIn);
    if(file.open(QIODevice::ReadOnly)) {
        if(document.setContent(&file)) {
            return true ;
        } else {
            Utils::error(tr("Unable to parse XML"));
        }
    } else {
        Utils::error(QString(tr("Unable to load file.")));
    }
    return false;
}

bool Compare::loadRule(const QString &filePath)
{
    if(!loadFile(filePath)) {
        return false;
    }
    toCompare = new Regola(document, filePath);
    doCompare();
    return true;
}

bool Compare::loadRule(QByteArray dataIn)
{
    if(!loadData(dataIn)) {
        return false;
    }
    toCompare = new Regola(document, "");
    doCompare();
    return true;
}

void Compare::doCompare()
{
    _areDifferent = false ;
    ignoreText = ui->ignoreText->isChecked();
    compareOnlyElements = ui->ignoreNonElements->isChecked();
    showOldValues = ui->showOldValues->isChecked();

    ui->textBrowser->clear();
    resultText = "" ;
    if((NULL == original) || (NULL == toCompare)) {
        Utils::error(this, tr("Empty data, impossible to compare."));
        return ;
    }
    resultText = "<html><head>";
    resultText += CSS_TEXT;
    resultText += "</head><body>";
    recurse(original->getItems(), toCompare->getItems(), 0);
    resultText += "</body></html>";
    ui->textBrowser->setHtml(resultText);
}

void Compare::filterElements(QVector<Element*> &origElements, QVector<Element*> &destElements, const bool isFilter)
{
    // Filter elements
    int origElementsSize = origElements.size();
    for(int i = 0; i < origElementsSize ; ++i) {
        Element * el = origElements.at(i);
        if(!isFilter || (isFilter && (el->getType() == Element ::ET_ELEMENT))) {
            destElements.append(el);
        }
    }
}

bool Compare::compareTwoText(const QString &str1, const QString &str2)
{
    QString s1 = str1 ;
    QString s2 = str2 ;
    s1.replace("\r\n", "\n");
    s2.replace("\r\n", "\n");
    bool equalCond = (s1 == s2);
    return equalCond;
}

void Compare::recurse(QVector<Element*> &inputOrigElements, QVector<Element*> &inputCompareElements, const int indent)
{
    // Filter elements
    QVector<Element*> origElements, compareElements ;

    filterElements(inputOrigElements, origElements, compareOnlyElements);
    filterElements(inputCompareElements, compareElements, compareOnlyElements);

    int origElementsSize = origElements.size();
    int compareElementsSize = compareElements.size();
    for(int elementIndex = 0 ; elementIndex < origElementsSize ; elementIndex ++) {
        EDiffStates state;
        Element *orig = origElements.at(elementIndex);
        Element* compare = NULL ;
        if(elementIndex < compareElementsSize) {
            compare = compareElements.at(elementIndex);
            state = DF_EQUAL ; // unless found otherwise
            if(orig->getType() != compare->getType()) {
                state = DF_DIFFERENT;
                setDifference();
            } else {
                if(orig->getType() == Element::ET_ELEMENT) {
                    if(orig->tag() != compare->tag()) {
                        state = DF_DIFFERENT;
                        setDifference();
                    }
                } else if(orig->getType() == Element::ET_COMMENT) {
                    if(!compareTwoText(orig->getComment(), compare->getComment())) {
                        state = DF_DIFFERENT;
                        setDifference();
                    }
                } else if(orig->getType() == Element::ET_TEXT) {
                    if(!compareTwoText(orig->text, compare->text)) {
                        state = DF_DIFFERENT;
                        setDifference();
                    }
                } else if(orig->getType() == Element::ET_PROCESSING_INSTRUCTION) {
                    if(!compareTwoText(orig->getPIData(), compare->getPIData())
                            ||  !compareTwoText(orig->getPITarget(), compare->getPITarget())) {
                        state = DF_DIFFERENT;
                        setDifference();
                    }
                }
                // confronta gli attributi
            } // end of test on same type
        } else {
            state = DF_ADDED ;
            setDifference();
        }
        if(state == DF_EQUAL) {
            QString text1 = orig->getAsSimpleText(false);
            QString text2 = compare->getAsSimpleText(false);
            if(!ignoreText) {
                if(!compareTwoText(text1, text2)) {
                    state = DF_DIFFERENT ;
                    setDifference();
                }
            }

            dumpElement(indent, orig, state);
            if(orig->getType() == Element::ET_ELEMENT) {
                compareAttributes(orig, compare);
            }
            closeTag(orig, state);
            dumpText(text1, state, indent);
            recurse(orig->getItems(), compare->getItems(), indent + 1);
            endElement(indent, orig, state);
        } else if(state == DF_DIFFERENT) {
            recurseAll(orig, DF_DIFFERENT, indent + 1);
        } else {
            recurseAll(orig, DF_ADDED, indent + 1);
        }
    }
    for(int elementIndex = origElementsSize ; elementIndex < compareElementsSize ; elementIndex++) {
        Element *element = compareElements.at(elementIndex);
        recurseAll(element, DF_MISSING, indent + 1);
    }
}

void Compare::closeTag(Element *element, const EDiffStates state)
{
    if(element->getType() == Element::ET_ELEMENT) {
        resultText += QString("<span class='E%1'>&gt;</span><br/>").arg(state);
    }
}

void Compare::recurseAll(Element* element, const EDiffStates state, const int indent)
{
    dumpElement(indent, element, state);
    if(element->getType() == Element::ET_ELEMENT) {
        dumpAttributes(element, state);
    }
    closeTag(element, state);
    QString text = element->getAsSimpleText(false);
    dumpText(text, state, indent);
    QVector<Element*> elements = element->getItems();
    int elementsSize = elements.size();
    for(int elementIndex = 0 ; elementIndex < elementsSize ; elementIndex ++) {
        Element *orig = elements.at(elementIndex);
        recurseAll(orig, state, indent + 1);
    }
    endElement(indent, element, state);
}

void Compare::dumpText(const QString text, const EDiffStates state, const int indent)
{
    if(!text.isEmpty()) {
        for(int i = 0 ; i < indent ; i++) {
            resultText += "&nbsp;";
        }
        if(state == DF_DIFFERENT) {
            resultText += "&nbsp;<span class='T1'>";
        } else {
            resultText += "&nbsp;<span class='T0'>";
        }
        resultText += convertTextInHTML(text);
        resultText += "</span><br/>";
    }
}

void Compare::dumpElement(const int indent, Element *element, const EDiffStates state)
{
    for(int i = 0 ; i < indent ; i++) {
        resultText += "&nbsp;";
    }
    switch(element->getType()) {
    case Element::ET_COMMENT:
        resultText += QString("<span class='C%1'>&lt;--%2--&gt;</span><br/>").arg(state).arg(element->getComment());
        break;
    case Element::ET_ELEMENT:
        resultText += QString("<span class='E%1'>&lt;%2</span>").arg(state).arg(element->tag());
        break;
    case Element::ET_PROCESSING_INSTRUCTION:
        resultText += QString("<span class='C%1'>&lt;?%2 %3?&gt;</span><br/>").arg(state).arg(element->getPITarget()).arg(element->getPIData());
        break;
    case Element::ET_TEXT:
        resultText += QString("<span class='T%1'>%1</span><br/>").arg(state).arg(convertTextInHTML(element->text));
        break;

    }
}

void Compare::endElement(const int indent, Element *element, const EDiffStates state)
{
    switch(element->getType()) {
    case Element::ET_TEXT:
        break;
    case Element::ET_COMMENT:
        break;
    case Element::ET_ELEMENT:
        for(int i = 0 ; i < indent ; i++) {
            resultText += "&nbsp;";
        }
        resultText += QString("<span class='E%1'>&lt;/%2&gt;</span><br/>").arg(state).arg(element->tag());
        break;
    case Element::ET_PROCESSING_INSTRUCTION:
        break;
    }
}


void Compare::insertAttrInHash(QHash<QString, Attribute*> &hash, Element *target)
{
    QVector<Attribute*>::iterator it;
    for(it = target->attributes.begin(); it != target->attributes.end(); ++it) {
        Attribute* attr = *it;
        hash[attr->name] = attr ;
    }
}

void Compare::compareAttributes(Element *orig, Element *compare)
{
    QHash<QString, Attribute*> hashCompare ;
    insertAttrInHash(hashCompare, compare);
    QVector<Attribute*>::iterator it;
    for(it = orig->attributes.begin(); it != orig->attributes.end(); ++it) {
        Attribute* attr = *it;
        if(hashCompare.contains(attr->name)) {
            Attribute *two = hashCompare[attr->name];
            if(attr->value == two->value) {
                dumpAttr(attr, DF_EQUAL);
            } else {
                dumpAttr(attr, DF_DIFFERENT);
                setDifference();
            }
            hashCompare.remove(attr->name);
        } else {
            dumpAttr(attr, DF_ADDED);
            setDifference();
        }
    }
    QHashIterator<QString, Attribute*> itt(hashCompare);
    while(itt.hasNext()) {
        itt.next();
        dumpAttr(itt.value(), DF_MISSING);
        setDifference();
    }
}

void Compare::dumpAttributes(Element *element, const EDiffStates state)
{
    QVector<Attribute*>::iterator it;
    for(it = element->attributes.begin(); it != element->attributes.end(); ++it) {
        Attribute* attr = *it;
        dumpAttr(attr, state);
    }
}

QString Compare::convertTextInHTML(const QString strIn)
{
    QString res ;
    int len = strIn.length();
    for(int i = 0 ; i < len ; i ++) {
        QChar ch = strIn.at(i);
        if(ch.isDigit() || (('a' <= ch) && (ch <= 'z')) || ((ch >= 'A') && (ch <= 'Z'))) {
            res.append(ch) ;
        } else {
            ushort valueInt = ch.unicode();
            res.append(QString("&#%1;"). arg(valueInt));
        }
    }
    return res;
}

void Compare::dumpAttr(Attribute *attr, const EDiffStates state)
{
    if(state == DF_EQUAL) {
        resultText += QString(" <span class='A%1'>%2=&quot;<span class='AV'>%3</span>&quot;</span>").arg(state).arg(attr->name).arg(convertTextInHTML(attr->value));
    } else {
        resultText += QString(" <span class='A%1'>%2=&quot;%3&quot;</span>").arg(state).arg(attr->name).arg(convertTextInHTML(attr->value));
    }

}

void Compare::accept()
{
    doCompare();
}

void Compare::on_copyToClipboard_clicked()
{
    QClipboard *clipboard = QApplication::clipboard();
    clipboard->setText(resultText);
}

void Compare::setDifference()
{
    _areDifferent = true ;
}
